import logging
import multiprocessing as mp
from .resource import *
from ...utils import Queue

__all__ = ['DistributedResourceManager', 'NodeResourceManager']

logger = logging.getLogger(__name__)

class DistributedResourceManager(object):
    LOCK = mp.Lock()
    REQUESTING_STACK = []
    MAX_CPU_COUNT = 0
    MAX_GPU_COUNT = 0
    NODE_RESOURCE_MANAGER = {}
    __instance = None
    def __new__(cls):
        # Singleton
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
        return cls.__instance

    @classmethod
    def add_remote(cls, remotes):
        """Enables dynamically removing nodes
        """
        remotes = remotes if isinstance(remotes, list) else [remotes]
        for remote in remotes:
            cls.NODE_RESOURCE_MANAGER[remote] = NodeResourceManager(remote)
        cls._refresh_resource()

    @classmethod
    def reserve_resource(cls, remote, resource):
        node_manager = cls.NODE_RESOURCE_MANAGER[remote]
        with cls.LOCK:
            if not node_manager.check_availability(resource):
                return False
            node_manager._request(remote, resource)
        logger.info('Reserved {} in {}'.format(resource, remote))
        return True

    @classmethod
    def release_reserved_resource(cls, remote, resource):
        node_manager = cls.NODE_RESOURCE_MANAGER[remote]
        node_manager._release(resource)
        cls._evoke_request()

    @classmethod
    def _refresh_resource(cls):
        cls.MAX_CPU_COUNT = max([x.get_all_resources()[0] for x in cls.NODE_RESOURCE_MANAGER.values()])
        cls.MAX_GPU_COUNT = max([x.get_all_resources()[1] for x in cls.NODE_RESOURCE_MANAGER.values()])

    @classmethod
    def _request(cls, resource):
        """ResourceManager, we recommand using scheduler instead of creating your own
        resource manager.
        """
        assert cls.check_possible(resource), \
            'Requested num_cpu={} and num_gpu={} should be less than or equal to' + \
            'largest node availability CPUs={}, GPUs={}'. \
            format(resource.num_cpus, resource.num_gpus, cls.MAX_GPU_COUNT, cls.MAX_CPU_COUNT)
       
        with cls.LOCK:
            node = cls.check_availability(resource)
            if node is not None:
                cls.NODE_RESOURCE_MANAGER[node]._request(node, resource)
                return

        logger.debug('Appending {} to Request Stack'.format(resource))
        request_semaphore = mp.Semaphore(0)
        with cls.LOCK:
            cls.REQUESTING_STACK.append((resource, request_semaphore))
        request_semaphore.acquire()
        return

    @classmethod
    def _release(cls, resource):
        logger.debug('\nReleasing resource {}'.format(resource))
        cls.NODE_RESOURCE_MANAGER[resource.node]._release(resource)
        cls._evoke_request()

    @classmethod
    def _evoke_request(cls):
        succeed = False
        with cls.LOCK:
            if len(cls.REQUESTING_STACK) > 0:
                resource, request_semaphore = cls.REQUESTING_STACK.pop()
                node = cls.check_availability(resource)
                if node is not None:
                    cls.NODE_RESOURCE_MANAGER[node]._request(node, resource)
                    logger.debug('\nEvoking requesting resource {}'.format(resource))
                    request_semaphore.release()
                    succeed = True
                else:
                    cls.REQUESTING_STACK.append((resource, request_semaphore))
                    return
        if succeed:
            cls._evoke_request()

    @classmethod
    def check_availability(cls, resource):
        """Unsafe check
        """
        candidate_nodes = cls._get_possible_nodes(resource)
        for node in candidate_nodes:
            if cls.NODE_RESOURCE_MANAGER[node].check_availability(resource):
                #logger.debug('\nSuccessfully find node {}'.format(node))
                return node
        return None

    @classmethod
    def check_possible(cls, resource):
        assert isinstance(resource, DistributedResource), \
            'Only support autogluon.resource.DistributedResource'
        if resource.num_cpus > cls.MAX_CPU_COUNT or resource.num_gpus > cls.MAX_GPU_COUNT:
            return False
        return True

    @classmethod
    def remove_remote(cls, remotes):
        #TODO 
        """Enables dynamically removing nodes
        """
        cls._refresh_resource()
        pass

    @classmethod
    def _get_possible_nodes(cls, resource):
        candidates = []
        for remote, manager in cls.NODE_RESOURCE_MANAGER.items():
            if manager.check_possible(resource):
                candidates.append(remote)
        return candidates

    def __repr__(self):
        reprstr = self.__class__.__name__ + '{\n'
        for remote, manager in self.NODE_RESOURCE_MANAGER.items():
            reprstr += '(Remote: {}, Resource: {})\n'.format(remote, manager)
        reprstr += '}'
        return reprstr


class NodeResourceManager(object):
    """Remote Resource Manager to keep track of the cpu and gpu usage
    """
    def __init__(self, remote):
        self.LOCK = mp.Lock()
        self.MAX_CPU_COUNT = get_remote_cpu_count(remote)
        self.MAX_GPU_COUNT = get_remote_gpu_count(remote)
        self.CPU_QUEUE = Queue()
        self.GPU_QUEUE = Queue()
        for cid in range(self.MAX_CPU_COUNT):
            self.CPU_QUEUE.put(cid)
        for gid in range(self.MAX_GPU_COUNT):
            self.GPU_QUEUE.put(gid)

    def _request(self, remote, resource):
        """ResourceManager, we recommand using scheduler instead of creating your own
        resource manager.
        """
        assert self.check_possible(resource), \
            'Requested num_cpu={} and num_gpu={} should be less than or equal to' + \
            'system availability CPUs={}, GPUs={}'. \
            format(resource.num_cpus, resource.num_gpus, self.MAX_GPU_COUNT, self.MAX_CPU_COUNT)

        with self.LOCK:
            cpu_ids = [self.CPU_QUEUE.get() for i in range(resource.num_cpus)]
            gpu_ids = [self.GPU_QUEUE.get() for i in range(resource.num_gpus)]
            resource._ready(remote, cpu_ids, gpu_ids)
            #logger.debug("\nReqeust succeed {}".format(resource))
            return
 
    def _release(self, resource):
        cpu_ids = resource.cpu_ids
        gpu_ids = resource.gpu_ids
        resource._release()
        if len(cpu_ids) > 0:
            for cid in cpu_ids:
                self.CPU_QUEUE.put(cid)
        if len(gpu_ids) > 0:
            for gid in gpu_ids:
                self.GPU_QUEUE.put(gid)

    def get_all_resources(self):
        return self.MAX_CPU_COUNT, self.MAX_GPU_COUNT

    def check_availability(self, resource):
        """Unsafe check
        """
        if resource.num_cpus > self.CPU_QUEUE.qsize() or resource.num_gpus > self.GPU_QUEUE.qsize():
            return False
        return True

    def check_possible(self, resource):
        assert isinstance(resource, DistributedResource), 'Only support autogluon.resource.Resources'
        if resource.num_cpus > self.MAX_CPU_COUNT or resource.num_gpus > self.MAX_GPU_COUNT:
            return False
        return True

    def __repr__(self):
        reprstr = self.__class__.__name__ + '(' + \
            '{} CPUs, '.format(self.MAX_CPU_COUNT) + \
            '{} GPUs)'.format(self.MAX_GPU_COUNT)
        return reprstr




import pandas as pd

from .utils import warning_filter

with warning_filter():
    from .utils.loaders import load_pd

__all__ = ['TabularDataset']


class TabularDataset(pd.DataFrame):
    """
    A dataset in tabular format (with rows = samples, columns = features/variables). 
    This object is essentially a pandas DataFrame (with some extra attributes) and all existing pandas methods can be applied to it. 
    For full list of methods/attributes, see pandas Dataframe documentation: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html

    Parameters
    ----------
    data : :class:`pd.DataFrame` or str
        If str, path to data file (CSV or Parquet format).
        If you already have your data in a :class:`pd.DataFrame`, you can specify it here.

    Attributes
    ----------
    file_path: (str)
        Path to data file from which this `TabularDataset` was created.
        None if `data` was a :class:`pd.DataFrame`.
    
    Note: In addition to these attributes, `TabularDataset` also shares all the same attributes and methods of a pandas Dataframe. 
    For a detailed list, see:  https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html

    Examples
    --------
    >>> from autogluon.core.dataset import TabularDataset
    >>> train_data = TabularDataset('https://autogluon.s3.amazonaws.com/datasets/Inc/train.csv')
    >>> train_data.head(30)
    >>> train_data.columns
    """

    _metadata = ['file_path']  # preserved properties that will be copied to a new instance of TabularDataset

    @property
    def _constructor(self):
        return TabularDataset

    @property
    def _constructor_sliced(self):
        return pd.Series

    def __init__(self, data, **kwargs):
        if isinstance(data, str):
            file_path = data
            data = load_pd.load(file_path)
        else:
            file_path = None
        super().__init__(data, **kwargs)
        self.file_path = file_path



"""Setup utils for autogluon. Only used for installing the code via setup.py, do not import after installation."""

# Refer to https://github.com/scikit-learn/scikit-learn/blob/main/sklearn/_min_dependencies.py for original implementation

import os

AUTOGLUON = 'autogluon'

AUTOGLUON_ROOT_PATH = os.path.abspath(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..', '..', '..')
)

PYTHON_REQUIRES = '>=3.6, <3.9'

# Only put packages here that would otherwise appear multiple times across different module's setup.py files.
DEPENDENT_PACKAGES = {
    'numpy': '==1.19.5',  # TODO: v0.3 consider upgrading
    'pandas': '>=1.0.0,<2.0',
    'scikit-learn': '>=0.23.2,<0.25',  # 0.22 crashes during efficient OOB in Tabular
    'scipy': '>=1.5.4,<1.7',
    'gluoncv': '>=0.10.1.post0,<0.11',
    'tqdm': '>=4.38.0',
    'Pillow': '<=8.1',
    'graphviz': '<0.9.0,>=0.8.1',
}
DEPENDENT_PACKAGES = {package: package + version for package, version in DEPENDENT_PACKAGES.items()}
# TODO: Use DOCS_PACKAGES and TEST_PACKAGES
DOCS_PACKAGES = []
TEST_PACKAGES = [
    'openml',
    'flake8',
    'pytest',
]


def load_version_file():
    with open(os.path.join(AUTOGLUON_ROOT_PATH, 'VERSION')) as version_file:
        version = version_file.read().strip()
    return version


def get_dependency_version_ranges(packages: list) -> list:
    return [package if package not in DEPENDENT_PACKAGES else DEPENDENT_PACKAGES[package] for package in packages]


def update_version(version, use_file_if_exists=True, create_file=False):
    """
    To release a new stable version on PyPi, simply tag the release on github, and the Github CI will automatically publish
    a new stable version to PyPi using the configurations in .github/workflows/pypi_release.yml .
    You need to increase the version number after stable release, so that the nightly pypi can work properly.
    """
    try:
        if not os.getenv('RELEASE'):
            from datetime import date
            minor_version_file_path = os.path.join(AUTOGLUON_ROOT_PATH, 'VERSION.minor')
            if use_file_if_exists and os.path.isfile(minor_version_file_path):
                with open(minor_version_file_path) as f:
                    day = f.read().strip()
            else:
                today = date.today()
                day = today.strftime("b%Y%m%d")
            version += day
    except Exception:
        pass
    if create_file and not os.getenv('RELEASE'):
        with open(os.path.join(AUTOGLUON_ROOT_PATH, 'VERSION.minor'), 'w') as f:
            f.write(day)
    return version


def create_version_file(*, version, submodule):
    print('-- Building version ' + version)
    if submodule is not None:
        version_path = os.path.join(AUTOGLUON_ROOT_PATH, submodule, 'src', AUTOGLUON, submodule, 'version.py')
    else:
        version_path = os.path.join(AUTOGLUON_ROOT_PATH, AUTOGLUON, 'src', AUTOGLUON, 'version.py')
    with open(version_path, 'w') as f:
        f.write(f'"""This is the {AUTOGLUON} version file."""\n')
        f.write("__version__ = '{}'\n".format(version))


def default_setup_args(*, version, submodule):
    from setuptools import find_packages
    long_description = open(os.path.join(AUTOGLUON_ROOT_PATH, 'README.md')).read()
    if submodule is None:
        name = AUTOGLUON
    else:
        name = f'{AUTOGLUON}.{submodule}'
    setup_args = dict(
        name=name,
        version=version,
        author='AutoGluon Community',
        url='https://github.com/awslabs/autogluon',
        description='AutoML for Text, Image, and Tabular Data',
        long_description=long_description,
        long_description_content_type='text/markdown',
        license='Apache-2.0',

        # Package info
        packages=find_packages('src'),
        package_dir={'': 'src'},
        namespace_packages=[AUTOGLUON],
        zip_safe=True,
        include_package_data=True,
        python_requires=PYTHON_REQUIRES,
        package_data={AUTOGLUON: [
            'LICENSE',
        ]},
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Education",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Intended Audience :: Customer Service",
            "Intended Audience :: Financial and Insurance Industry",
            "Intended Audience :: Healthcare Industry",
            "Intended Audience :: Telecommunications Industry",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: MacOS",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            'Programming Language :: Python :: 3',
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Information Analysis",
            "Topic :: Scientific/Engineering :: Image Recognition",
        ],
        project_urls={
            'Documentation': 'https://auto.gluon.ai',
            'Bug Reports': 'https://github.com/awslabs/autogluon/issues',
            'Source': 'https://github.com/awslabs/autogluon/',
            'Contribute!': 'https://github.com/awslabs/autogluon/blob/master/CONTRIBUTING.md',
        },
    )
    return setup_args
